@@ -1,7 +1,4 @@
 """Top-level package for PyGeoOGC."""
-import asyncio
-import sys
-
 from pkg_resources import DistributionNotFound, get_distribution
 
 from .exceptions import (
@@ -9,18 +6,16 @@ from .exceptions import (
     InvalidInputValue,
     MissingInputs,
     ServerError,
+    ServiceError,
     ThreadingException,
     ZeroMatched,
 )
 from .print_versions import show_versions
 from .pygeoogc import WFS, WMS, ArcGISRESTful, ServiceURL
-from .utils import MatchCRS, RetrySession, async_requests
+from .utils import MatchCRS, RetrySession
 
 try:
     __version__ = get_distribution(__name__).version
 except DistributionNotFound:
     # package is not installed
     pass
-
-if sys.platform.startswith("win"):
-    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
