@@ -19,13 +19,30 @@ class ServerError(Exception):
         return self.message
 
 
+class ServiceError(Exception):
+    """Exception raised when the requested data is not available on the server.
+
+    Parameters
+    ----------
+    url : str
+        The server url
+    """
+
+    def __init__(self, url: str) -> None:
+        self.message = f"Service is currently not available, try again later:\n{url}"
+        super().__init__(self.message)
+
+    def __str__(self) -> str:
+        return self.message
+
+
 class ThreadingException(Exception):
     """Exception raised when the requested data is not available on the server.
 
     Parameters
     ----------
     itr : int
-        The number of iteration where the exception occured
+        The number of iteration where the exception occurred
     msg : str
         The exception error message
     """
@@ -39,7 +56,20 @@ class ThreadingException(Exception):
 
 
 class ZeroMatched(ValueError):
-    """Exception raised when a function argument is missing."""
+    """Exception raised when a function argument is missing.
+
+    Parameters
+    ----------
+    msg : str
+        The exception error message
+    """
+
+    def __init__(self, msg: str = "") -> None:
+        self.message = f"Query returned no feature ID. {msg}"
+        super().__init__(self.message)
+
+    def __str__(self) -> str:
+        return self.message
 
 
 class InvalidInputValue(Exception):
