@@ -1,40 +1,38 @@
 """Some utilities for PyGeoOGC."""
-import asyncio
 import math
 import socket
 from concurrent import futures
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    Iterable,
-    List,
-    Mapping,
-    MutableMapping,
-    Optional,
-    Tuple,
-    Union,
-)
+from dataclasses import dataclass
+from pathlib import Path
+from sqlite3 import OperationalError
+from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Union
 from unittest.mock import _patch, patch
 
-import aiohttp
-import cytoolz as tlz
 import defusedxml.ElementTree as etree
-import nest_asyncio
-import orjson as json
 import pyproj
-from requests import Response, Session
+import simplejson as json
+from requests import Response
 from requests.adapters import HTTPAdapter
 from requests.exceptions import RequestException
+from requests_cache import CachedSession
+from requests_cache.backends.sqlite import DbCache
 from shapely import ops
-from shapely.geometry import LineString, MultiPoint, MultiPolygon, Point, Polygon, box
+from shapely.geometry import (
+    LineString,
+    MultiLineString,
+    MultiPoint,
+    MultiPolygon,
+    Point,
+    Polygon,
+    box,
+)
 from urllib3 import Retry
 
-from .exceptions import InvalidInputType, InvalidInputValue, ThreadingException, ZeroMatched
+from .exceptions import InvalidInputType, ThreadingException, ZeroMatched
 
-nest_asyncio.apply()
 DEF_CRS = "epsg:4326"
 BOX_ORD = "(west, south, east, north)"
+EXPIRE = 24 * 60 * 60
 
 
 class RetrySession:
@@ -54,6 +52,9 @@ class RetrySession:
         A tuple of status codes that trigger the reply behaviour, defaults to (500, 502, 504).
     prefixes : tuple, optional
         The prefixes to consider, defaults to ("http://", "https://")
+    cache_name : str, optional
+        Path to a folder for caching the session, default to None which uses
+        system's temp directory.
     """
 
     def __init__(
@@ -62,10 +63,19 @@ class RetrySession:
         backoff_factor: float = 0.3,
         status_to_retry: Tuple[int, ...] = (500, 502, 504),
         prefixes: Tuple[str, ...] = ("https://",),
+        cache_name: Optional[Union[str, Path]] = None,
     ) -> None:
-        self.session = Session()
-        self.retries = retries
+        self.cache_name = (
+            Path("cache", "http_cache.sqlite") if cache_name is None else Path(cache_name)
+        )
+        try:
+            backend = DbCache(self.cache_name, fast_save=True, timeout=2)
+            self.session = CachedSession(expire_after=EXPIRE, backend=backend)
+            self.session.remove_expired_responses()
+        except OperationalError:
+            self._backup_db()
 
+        self.retries = retries
         retry_args = {
             "total": retries,
             "read": retries,
@@ -88,26 +98,39 @@ class RetrySession:
         self,
         url: str,
         payload: Optional[Mapping[str, Any]] = None,
-        headers: Optional[MutableMapping[str, Any]] = None,
+        headers: Optional[Mapping[str, Any]] = None,
     ) -> Response:
         """Retrieve data from a url by GET and return the Response."""
         try:
             return self.session.get(url, params=payload, headers=headers)
+        except OperationalError:
+            self._backup_db()
+            return self.session.get(url, params=payload, headers=headers)
         except (ConnectionError, RequestException):
             raise ConnectionError(f"Connection failed after {self.retries} retries.")
 
     def post(
         self,
         url: str,
-        payload: Optional[MutableMapping[str, Any]] = None,
-        headers: Optional[MutableMapping[str, Any]] = None,
+        payload: Optional[Mapping[str, Any]] = None,
+        headers: Optional[Mapping[str, Any]] = None,
     ) -> Response:
         """Retrieve data from a url by POST and return the Response."""
         try:
             return self.session.post(url, data=payload, headers=headers)
+        except OperationalError:
+            self._backup_db()
+            return self.session.post(url, data=payload, headers=headers)
         except (ConnectionError, RequestException):
             raise ConnectionError(f"Connection failed after {self.retries} retries.")
 
+    def _backup_db(self) -> None:
+        """Use a backup database if the current database is locked."""
+        self.cache_name = Path(f"{self.cache_name.parent}/{self.cache_name.stem}_back_up.sqlite")
+        backend = DbCache(self.cache_name, fast_save=True, timeout=2)
+        self.session = CachedSession(expire_after=EXPIRE, backend=backend)
+        self.session.remove_expired_responses()
+
     @staticmethod
     def onlyipv4() -> _patch:
         """Disable IPv6 and only use IPv4."""
@@ -141,156 +164,6 @@ class RetrySession:
         return patch("socket.getaddrinfo", side_effect=getaddrinfo_ipv4)
 
 
-async def _request_binary(
-    url: str,
-    session_req: aiohttp.ClientSession,
-    payload: Dict[str, Optional[MutableMapping[str, Any]]],
-) -> bytes:
-    """Create an async request and return the response as binary.
-
-    Parameters
-    ----------
-    url : str
-        URL to be retrieved
-    session_req : ClientSession
-        A ClientSession for sending the request
-    paylod: dict
-        The request payload
-
-    Returns
-    -------
-    bytes
-        The retrieved response as binary
-    """
-    async with session_req(url, **payload) as response:
-        return await response.read()
-
-
-async def _request_json(
-    url: str,
-    session_req: aiohttp.ClientSession,
-    payload: Dict[str, Optional[MutableMapping[str, Any]]],
-) -> MutableMapping[str, Any]:
-    """Create an async request and return the response as json.
-
-    Parameters
-    ----------
-    url : str
-        URL to be retrieved
-    session_req : ClientSession
-        A ClientSession for sending the request
-    paylod: dict
-        The request payload
-
-    Returns
-    -------
-    dict
-        The retrieved response as json
-    """
-    async with session_req(url, **payload) as response:
-        return await response.json()
-
-
-async def _request_text(
-    url: str,
-    session_req: aiohttp.ClientSession,
-    payload: Dict[str, Optional[MutableMapping[str, Any]]],
-) -> str:
-    """Create an async request and return the response as string.
-
-    Parameters
-    ----------
-    url : str
-        URL to be retrieved
-    session_req : ClientSession
-        A ClientSession for sending the request
-    paylod: dict
-        The request payload
-
-    Returns
-    -------
-    dict
-        The retrieved response as string
-    """
-    async with session_req(url, **payload) as response:
-        return await response.text()
-
-
-async def _async_session(
-    url_payload: Tuple[Tuple[str, Optional[MutableMapping[str, Any]]], ...],
-    read: str,
-    request: str,
-) -> Callable:
-    """Create an async session for sending requests.
-
-    Parameters
-    ----------
-    url_payload : list of tuples of urls and payloads
-        A list of URLs or URLs with their payloads to be retrieved.
-    read : str
-        The method for returning the request; binary, json, and text.
-    request : str
-        The request type; GET or POST.
-
-    Returns
-    -------
-    asyncio.gather
-        An async gather function
-    """
-    async with aiohttp.ClientSession(json_serialize=json.dumps) as session:
-        read_method = {"binary": _request_binary, "json": _request_json, "text": _request_text}
-        if read not in read_method:
-            raise InvalidInputValue("read", list(read_method.keys()))
-
-        request_method = {"GET": session.get, "POST": session.post}
-        if request not in request_method:
-            raise InvalidInputValue("method", list(request_method.keys()))
-
-        paylod = {"GET": "params", "POST": "data"}
-        tasks = (
-            read_method[read](u, request_method[request], {paylod[request]: p})
-            for u, p in url_payload
-        )
-        return await asyncio.gather(*tasks, return_exceptions=True)  # type: ignore
-
-
-def async_requests(
-    url_payload: List[Tuple[str, Optional[MutableMapping[str, Any]]]],
-    read: str,
-    request: str = "GET",
-    max_workers: int = 8,
-) -> List[Union[str, MutableMapping[str, Any], bytes]]:
-    """Send async requests.
-
-    This function is based on
-    `this <https://github.com/HydrologicEngineeringCenter/data-retrieval-scripts/blob/master/qpe_async_download.py>`__
-    script.
-
-    Parameters
-    ----------
-    url_payload : list of tuples
-        A list of URLs and payloads as a tuple.
-    read : str
-        The method for returning the request; binary, json, and text.
-    request : str, optional
-        The request type; GET or POST, defaults to GET.
-    max_workers : int, optional
-        The maximum number of async processes, defaults to 8.
-
-    Returns
-    -------
-    list
-        A list of responses
-    """
-    chunked_urls = tlz.partition_all(max_workers, url_payload)
-
-    results = (
-        asyncio.get_event_loop().run_until_complete(_async_session(c, read, request))
-        for c in chunked_urls
-    )
-    return list(tlz.concat(results))
-
-
 def threading(
     func: Callable,
     iter_list: Iterable,
@@ -397,12 +270,13 @@ def traverse_json(
     return []
 
 
+@dataclass
 class ESRIGeomQuery:
     """Generate input geometry query for ArcGIS RESTful services.
 
     Parameters
     ----------
-    geometry : tuple or Polygon
+    geometry : tuple or Polygon or Point or LineString
         The input geometry which can be a point (x, y), a list of points [(x, y), ...],
         bbox (xmin, ymin, xmax, ymax), or a Shapely's Polygon.
     wkid : int
@@ -412,18 +286,13 @@ class ESRIGeomQuery:
         for reference.
     """
 
-    def __init__(
-        self,
-        geometry: Union[
-            Tuple[float, float],
-            List[Tuple[float, float]],
-            Tuple[float, float, float, float],
-            Polygon,
-        ],
-        wkid: int,
-    ) -> None:
-        self.geometry = geometry
-        self.wkid = wkid
+    geometry: Union[
+        Tuple[float, float],
+        List[Tuple[float, float]],
+        Tuple[float, float, float, float],
+        Polygon,
+    ]
+    wkid: int
 
     def point(self) -> Dict[str, Union[str, bytes]]:
         """Query for a point."""
@@ -455,12 +324,21 @@ class ESRIGeomQuery:
     def polygon(self) -> Dict[str, Union[str, bytes]]:
         """Query for a polygon."""
         if not isinstance(self.geometry, Polygon):
-            raise InvalidInputType("geomtry", "Shapely's Polygon")
+            raise InvalidInputType("geometry", "Shapely's Polygon")
 
         geo_type = "esriGeometryPolygon"
         geo_json = {"rings": [[[x, y] for x, y in zip(*self.geometry.exterior.coords.xy)]]}
         return self.get_payload(geo_type, geo_json)
 
+    def polyline(self) -> Dict[str, Union[str, bytes]]:
+        """Query for a polyline."""
+        if not isinstance(self.geometry, LineString):
+            raise InvalidInputType("geometry", "Shapely's LineString")
+
+        geo_type = "esriGeometryPolyline"
+        geo_json = {"paths": [[[x, y] for x, y in zip(*self.geometry.coords.xy)]]}
+        return self.get_payload(geo_type, geo_json)
+
     def get_payload(self, geo_type: str, geo_json: Dict[str, Any]) -> Dict[str, Union[str, bytes]]:
         """Generate a request payload based on ESRI template.
 
@@ -470,6 +348,11 @@ class ESRIGeomQuery:
             Type of the input geometry
         geo_json : dict
             Geometry in GeoJson format.
+
+        Returns
+        -------
+        dict
+            An ESRI geometry payload.
         """
         esri_json = json.dumps({**geo_json, "spatialRelference": {"wkid": str(self.wkid)}})
         return {
@@ -480,56 +363,104 @@ class ESRIGeomQuery:
 
 
 class MatchCRS:
-    """Match CRS of a input geometry (Polygon, bbox, coord) with the output CRS.
+    """Reproject a geometry to another CRS.
 
     Parameters
     ----------
-    geometry : tuple or Polygon
-        The input geometry (Polygon, bbox, coord)
     in_crs : str
-        The spatial reference of the input geometry
+        Spatial reference of the input geometry
     out_crs : str
-        The target spatial reference
+        Target spatial reference
     """
 
-    @staticmethod
+    def __init__(self, in_crs: str, out_crs: str):
+        self.project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
+
     def geometry(
-        geom: Union[Polygon, MultiPolygon, Point, MultiPoint], in_crs: str, out_crs: str
-    ) -> Union[Polygon, MultiPolygon, Point, MultiPoint]:
-        """Reproject a geometry to the specified output CRS."""
-        if not isinstance(geom, (Polygon, MultiPolygon, Point, MultiPoint)):
-            raise InvalidInputType("geom", "Polygon, MultiPolygon, Point, or MultiPoint")
+        self, geom: Union[Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint]
+    ) -> Union[Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint]:
+        """Reproject a geometry to the specified output CRS.
 
-        project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
-        return ops.transform(project, geom)
+        Parameters
+        ----------
+        geom : LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint
+            Input geometry.
 
-    @staticmethod
-    def bounds(
-        geom: Tuple[float, float, float, float], in_crs: str, out_crs: str
-    ) -> Tuple[float, float, float, float]:
-        """Reproject a bounding box to the specified output CRS."""
+        Returns
+        -------
+        LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint
+            Input geometry in the specified CRS.
+
+        Examples
+        --------
+        >>> from pygeoogc import MatchCRS
+        >>> from shapely.geometry import Point
+        >>> point = Point(-7766049.665, 5691929.739)
+        >>> MatchCRS("epsg:3857", "epsg:4326").geometry(point).xy
+        (array('d', [-69.7636111130079]), array('d', [45.44549114818127]))
+        """
+        if not isinstance(
+            geom, (Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint)
+        ):
+            types = "LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint"
+            raise InvalidInputType("geom", types)
+
+        return ops.transform(self.project, geom)
+
+    def bounds(self, geom: Tuple[float, float, float, float]) -> Tuple[float, float, float, float]:
+        """Reproject a bounding box to the specified output CRS.
+
+        Parameters
+        ----------
+        geom : tuple
+            Input bounding box (xmin, ymin, xmax, ymax).
+
+        Returns
+        -------
+        tuple
+            Input bounding box in the specified CRS.
+
+        Examples
+        --------
+        >>> from pygeoogc import MatchCRS
+        >>> bbox = (-7766049.665, 5691929.739, -7763049.665, 5696929.739)
+        >>> MatchCRS("epsg:3857", "epsg:4326").bounds(bbox)
+        (-69.7636111130079, 45.44549114818127, -69.73666165448431, 45.47699468552394)
+        """
         if not (isinstance(geom, tuple) and len(geom) == 4):
-            raise InvalidInputType("geom", "tuple of length 4", BOX_ORD)
+            raise InvalidInputType("geom", "tuple", BOX_ORD)
 
-        project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
-        bbox: Tuple[float, float, float, float] = ops.transform(project, box(*geom)).bounds
-        return bbox
+        return ops.transform(self.project, box(*geom)).bounds
 
-    @staticmethod
-    def coords(
-        geom: Tuple[Tuple[float, ...], Tuple[float, ...]], in_crs: str, out_crs: str
-    ) -> Tuple[Any, ...]:
-        """Reproject a list of coordinates to the specified output CRS."""
-        if not (isinstance(geom, tuple) and len(geom) == 2):
-            raise InvalidInputType("geom", "tuple of length 2", "((xs), (ys))")
+    def coords(self, geom: List[Tuple[float, float]]) -> List[Tuple[Any, ...]]:
+        """Reproject a list of coordinates to the specified output CRS.
 
-        project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
-        return tuple(zip(*[project(x, y) for x, y in zip(*geom)]))
+        Parameters
+        ----------
+        geom : list of tuple
+            Input coords [(x1, y1), ...].
+
+        Returns
+        -------
+        tuple
+            Input list of coords in the specified CRS.
+
+        Examples
+        --------
+        >>> from pygeoogc import MatchCRS
+        >>> coords = [(-7766049.665, 5691929.739)]
+        >>> MatchCRS("epsg:3857", "epsg:4326").coords(coords)
+        [(-69.7636111130079, 45.44549114818127)]
+        """
+        if not (isinstance(geom, list) and all(len(c) == 2 for c in geom)):
+            raise InvalidInputType("geom", "list of tuples", "[(x1, y1), ...]")
+
+        return list(zip(*self.project(*zip(*geom))))
 
 
 def check_bbox(bbox: Tuple[float, float, float, float]) -> None:
     """Check if an input inbox is a tuple of length 4."""
-    if not isinstance(bbox, tuple) or len(bbox) != 4:
+    if not (isinstance(bbox, tuple) and len(bbox) == 4):
         raise InvalidInputType("bbox", "tuple", BOX_ORD)
 
 
@@ -554,7 +485,7 @@ def bbox_resolution(
     """
     check_bbox(bbox)
 
-    bbox = MatchCRS.bounds(bbox, bbox_crs, DEF_CRS)
+    bbox = MatchCRS(bbox_crs, DEF_CRS).bounds(bbox)
     west, south, east, north = bbox
     geod = pyproj.Geod(ellps="WGS84")
 
@@ -593,7 +524,7 @@ def bbox_decompose(
         The first element is a list of bboxes and the second one is width of the last bbox
     """
     check_bbox(bbox)
-    _bbox = MatchCRS.bounds(bbox, box_crs, DEF_CRS)
+    _bbox = MatchCRS(box_crs, DEF_CRS).bounds(bbox)
     width, height = bbox_resolution(_bbox, resolution, DEF_CRS)
 
     n_px = width * height
@@ -627,9 +558,9 @@ def bbox_decompose(
 
             coords[-1] = (coords[-1][0], dest)
 
-            nd = len(coords)
-            divs = [dim for _ in range(nd)]
-            divs[-1] = px - (nd - 1) * dim
+            n_dim = len(coords)
+            divs = [dim for _ in range(n_dim)]
+            divs[-1] = px - (n_dim - 1) * dim
             mul -= 0.1
         return coords, divs
 
@@ -642,7 +573,7 @@ def bbox_decompose(
     bboxs = []
     for i, ((bottom, top), h) in enumerate(zip(lats, heights)):
         for j, ((left, right), w) in enumerate(zip(lons, widths)):
-            bx_crs = MatchCRS.bounds((left, bottom, right, top), DEF_CRS, box_crs)
+            bx_crs = MatchCRS(DEF_CRS, box_crs).bounds((left, bottom, right, top))
             bboxs.append((bx_crs, f"{i}_{j}", w, h))
     return bboxs
 
